from ..backend_object import BackendObject

class BoolResult(BackendObject):
    def __init__(self, op=None, args=None):
        self._op = op
        self._args = args

    def value(self):
        raise NotImplementedError()

    def __len__(self):
        return BackendError()

    def __eq__(self, other):
        raise NotImplementedError()

    def __and__(self, other):
        raise NotImplementedError()

    def __invert__(self):
        raise NotImplementedError()

    def __or__(self, other):
        raise NotImplementedError()

    def identical(self, other):
        if self.value != other.value:
            return False
        if self._op != other._op:
            return False
        if self._args != other._args:
            return False
        return True

    def union(self, other):
        raise NotImplementedError()

    def size(self): #pylint:disable=no-self-use
        return None

    @staticmethod
    def is_maybe(o):
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return isinstance(o, MaybeResult)

    @staticmethod
    def has_true(o):
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return o is True or (isinstance(o, BoolResult) and True in o.value)

    @staticmethod
    def has_false(o):
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return o is False or (isinstance(o, BoolResult) and False in o.value)

    @staticmethod
    def is_true(o):
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return o is True or (isinstance(o, TrueResult))

    @staticmethod
    def is_false(o):
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return o is False or (isinstance(o, FalseResult))

class TrueResult(BoolResult):
    cardinality = 1

    @property
    def value(self):
        return (True, )

    def identical(self, other):
        return isinstance(other, TrueResult)

    def __eq__(self, other):
        if isinstance(other, FalseResult):
            return FalseResult()
        elif isinstance(other, TrueResult):
            return TrueResult()
        else:
            return MaybeResult()

    def __invert__(self):
        return FalseResult()

    def __or__(self, other):
        return TrueResult()

    def __and__(self, other):
        if BoolResult.is_maybe(other):
            return MaybeResult()
        elif BoolResult.is_false(other):
            return FalseResult()
        else:
            return TrueResult()

    def union(self, other):
        if other is True or type(other) is TrueResult:
            return TrueResult()
        elif other is False or type(other) is FalseResult:
            return MaybeResult()
        elif type(other) is MaybeResult:
            return MaybeResult()
        else:
            return NotImplemented

    def __repr__(self):
        return '<True>'

    def __bool__(self):
        return True

class FalseResult(BoolResult):
    cardinality = 1

    @property
    def value(self):
        return (False, )

    def identical(self, other):
        return isinstance(other, FalseResult)

    def __eq__(self, other):
        if isinstance(other, FalseResult):
            return TrueResult()
        elif isinstance(other, TrueResult):
            return FalseResult()
        else:
            return MaybeResult()

    def __invert__(self):
        return TrueResult()

    def __and__(self, other):
        return FalseResult()

    def __or__(self, other):
        return other

    def __repr__(self):
        return '<False>'

    def union(self, other):
        if other is True or type(other) is TrueResult:
            return MaybeResult()
        elif other is False or type(other) is FalseResult:
            return FalseResult()
        elif type(other) is MaybeResult:
            return MaybeResult()
        else:
            return NotImplemented

    def __bool__(self):
        return False

class MaybeResult(BoolResult):
    cardinality = 2

    @property
    def value(self):
        return (True, False)

    def identical(self, other):
        return isinstance(other, MaybeResult)

    def __eq__(self, other):
        return MaybeResult()

    def __invert__(self):
        return MaybeResult()

    def __and__(self, other):
        if BoolResult.is_false(other):
            return FalseResult()
        else:
            return MaybeResult()

    def union(self, other):
        return MaybeResult()

    def __or__(self, other):
        if BoolResult.is_true(other):
            return TrueResult()
        else:
            return self

    def __repr__(self):
        if self._op is None:
            return '<Maybe>'
        else:
            return '<Maybe(%s, %s)>' % (self._op, self._args)

    def __bool__(self):
        return False



from ..errors import BackendError, ClaripyValueError
from ..ast.base import Base



#import time

from .backend import Backend
from .backend_z3 import BackendZ3
try:
    from . import remotetasks
except ImportError:
    remotetasks = None

class TrackingSolver:
    def __init__(self, sz3):
        self.sz3 = sz3
        self.constraints = []

    def plus_extra(self, extra_constraints):
        return self.constraints + list(extra_constraints)


def get(res):
    #before = time.time()
    result = res.get(interval=0.02)
    # print("task took %f" % (time.time() - before))
    return result

class BackendRemote(Backend):
    def __init__(self, host='localhost', port=1337, local_timeout=-1):
        if remotetasks is None:
            raise ImportError("celery/pymongo is required for BackendRemote")

        super(BackendRemote, self).__init__()
        self.bz3 = BackendZ3()
        self.local_timeout = local_timeout

    def solver(self, timeout=None):
        timeout = self.local_timeout if timeout is None else min(self.local_timeout, timeout)
        return TrackingSolver(self.bz3.solver(timeout=timeout))

    #def _abstract(self, e):
    #   return e

    def _convert(self, e):
        return e

    def resolve(self, ast):
        return ast

    def _add(self, s, c):
        s.constraints.extend(c)
        self.bz3.add(s.sz3, c)

    def _check(self, s, extra_constraints=()):
        for x in s.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        # print(pickle.dumps(s.plus_extra(extra_constraints)))
        res = remotetasks.check.delay(s.plus_extra(extra_constraints))
        return get(res)

    def _eval(self, expr, n, solver=None, extra_constraints=()):
        for x in solver.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        # print(pickle.dumps(s.plus_extra(extra_constraints)))
        res = remotetasks.eval.delay(solver.plus_extra(extra_constraints), expr, n)
        return get(res)

    def _batch_eval(self, exprs, n, extra_constraints=(), solver=None):
        for x in solver.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        res = remotetasks.batch_eval.delay(solver.plus_extra(extra_constraints), exprs, n)
        return get(res)

    def _results(self, s, extra_constraints=(), generic_model=True):
        for x in s.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        # print(pickle.dumps(s.plus_extra(extra_constraints)))
        res = remotetasks.results.delay(s.plus_extra(extra_constraints))
        return get(res)

    def _min(self, expr, solver=None, extra_constraints=()):
        for x in solver.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        # print(pickle.dumps(s.plus_extra(extra_constraints)))
        res = remotetasks.min.delay(solver.plus_extra(extra_constraints), expr)
        return get(res)

    def _max(self, expr, extra_constraints=(), solver=None):
        for x in solver.plus_extra(extra_constraints):
            if hasattr(x, 'make_uuid'):
                x.make_uuid()
        # print(pickle.dumps(s.plus_extra(extra_constraints)))
        res = remotetasks.max.delay(solver.plus_extra(extra_constraints), expr)
        return get(res)

    def _size(self, o):
        return o.length

    def _name(self, o):
        raise Exception('wat') # TODO: Be more explicit :p

    def _identical(self, a, b):
        return a.identical(b)



class BackendManager:
    def __init__(self):
        self._eager_backends = [ ]
        self._quick_backends = [ ]
        self._all_backends = [ ]
        self._backends_by_type = { }
        self._backends_by_name = { }

    def _register_backend(self, b, name, eager, quick):
        self._backends_by_name[name] = b
        self._backends_by_type[b.__class__.__name__] = b
        self._all_backends.append(b)
        if eager:
            self._eager_backends.append(b)

        if quick:
            self._quick_backends.append(b)

    def __getattr__(self, a):
        if a in self._backends_by_name:
            return self._backends_by_name[a]
        else:
            raise AttributeError(a)

    def downsize(self):
        for b in self._all_backends:
            b.downsize()

backends = BackendManager()



class SatCacheMixin:
    def __init__(self, *args, **kwargs):
        super(SatCacheMixin, self).__init__(*args, **kwargs)
        self._cached_satness = None

    def _blank_copy(self, c):
        super(SatCacheMixin, self)._blank_copy(c)
        c._cached_satness = None

    def _copy(self, c):
        super(SatCacheMixin, self)._copy(c)
        c._cached_satness = self._cached_satness

    def __getstate__(self):
        return self._cached_satness, super().__getstate__()

    def __setstate__(self, s):
        self._cached_satness, base_state = s
        super().__setstate__(base_state)

    #
    # SAT caching
    #

    def add(self, constraints, **kwargs):
        added = super(SatCacheMixin, self).add(constraints, **kwargs)
        if len(added) > 0 and any(c is false for c in added):
            self._cached_satness = False
        elif self._cached_satness is True:
            self._cached_satness = None
        return added

    def simplify(self):
        new_constraints = super(SatCacheMixin, self).simplify()
        if len(new_constraints) > 0 and any(c is false for c in new_constraints):
            self._cached_satness = False
        return new_constraints

    def satisfiable(self, extra_constraints=(), **kwargs):
        if self._cached_satness is False:
            return False
        if self._cached_satness is True and len(extra_constraints) == 0:
            return True
        r = super(SatCacheMixin, self).satisfiable(
            extra_constraints=extra_constraints, **kwargs
        )
        if len(extra_constraints) == 0:
            self._cached_satness = r
        return r

    def eval(self, e, n, extra_constraints=(), **kwargs):
        if self._cached_satness is False: raise UnsatError("cached unsat")
        try:
            r = super(SatCacheMixin, self).eval(
                e, n,
                extra_constraints=extra_constraints, **kwargs
            )
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def batch_eval(self, e, n, extra_constraints=(), **kwargs):
        if self._cached_satness is False: raise UnsatError("cached unsat")
        try:
            r = super(SatCacheMixin, self).batch_eval(
                e, n,
                extra_constraints=extra_constraints, **kwargs
            )
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def max(self, e, extra_constraints=(), **kwargs):
        if self._cached_satness is False: raise UnsatError("cached unsat")
        try:
            r = super(SatCacheMixin, self).max(
                e,
                extra_constraints=extra_constraints, **kwargs
            )
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def min(self, e, extra_constraints=(), **kwargs):
        if self._cached_satness is False: raise UnsatError("cached unsat")
        try:
            r = super(SatCacheMixin, self).min(
                e,
                extra_constraints=extra_constraints, **kwargs
            )
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

    def solution(self, e, v, extra_constraints=(), **kwargs):
        if self._cached_satness is False: raise UnsatError("cached unsat")
        try:
            r = super(SatCacheMixin, self).solution(
                e, v,
                extra_constraints=extra_constraints, **kwargs
            )
            self._cached_satness = True
            return r
        except UnsatError:
            if len(extra_constraints) == 0:
                self._cached_satness = False
            raise

from .. import false
from ..errors import UnsatError



#pylint:disable=redefined-outer-name
Bits = lambda *args, **kwargs: None
BV = lambda *args, **kwargs: None
VS = lambda *args, **kwargs: None
FP = lambda *args, **kwargs: None
Bool = lambda *args, **kwargs: None
Int = lambda *args, **kwargs: None
Base = lambda *args, **kwargs: None
true = lambda *args, **kwargs: None
false = lambda *args, **kwargs: None
String = lambda *args, **kwargs: None
all_operations = None

def _import():
    global Bits, BV, VS, FP, Bool, Int, Base, String, true, false, all_operations

    from .bits import Bits
    from .bv import BV
    from .vs import VS
    from .fp import FP
    from .bool import Bool, true, false
    from .int import Int
    from .base import Base
    from .strings import String
    from .. import ops as all_operations



from .light_frontend import LightFrontend
from .full_frontend import FullFrontend
from .hybrid_frontend import HybridFrontend
from .composite_frontend import CompositeFrontend
from .replacement_frontend import ReplacementFrontend
